/*	bb5.c

	Basic Block compiler in C:  Code Generator

	2017 by H. Dietz
*/

#include	"bb.h"

static	tuple	code = {
	&code, &code, UNDEF
};

static int
ttoi(register tuple *t)
{
	/* tuple pointer to int position */
	register int i = 0;
	register tuple *p = code.next;

	while (p != &code) {
		if (p == t) return(i);
		++i;
		p = p->next;
	}

	return(-1);
}

static void
cgbin(char *s, tuple *p)
{
	printf("%s(%d, %d)", s, ttoi(p->targ[0]), ttoi(p->targ[1]));
}

static tuple *
mktuple(void)
{
	/* make a tuple and link it into code */
	register tuple *p = ((tuple *) malloc(sizeof(tuple)));

	p->next = &code;
	(p->prev = code.prev)->next = p;
	return(code.prev = p);
}

static void
rmtuple(register tuple *p)
{
	/* remove this tuple */

	(p->prev)->next = p->next;
	(p->next)->prev = p->prev;
	free((char *) p);
}

tuple *
binop(opcode o, tuple *t1, tuple *t2)
{
	register tuple *p = mktuple();

	p->oarg = o;
	p->targ[0] = t1;
	p->targ[1] = t2;
	return(p);
}

tuple *
cop(konst c)
{
	register tuple *p = mktuple();

	p->oarg = CONST;
	p->carg = c;
	return(p);
}

tuple *
ldop(var *v)
{
	register tuple *p = mktuple();

	p->oarg = LD;
	p->varg = v;
	return(p);
}

tuple *
ldxop(var *v, tuple *t)
{
	register tuple *p = mktuple();

	p->oarg = LDX;
	p->varg = v;
	p->targ[0] = t;
	return(p);
}

tuple *
stop(var *v, tuple *t)
{
	register tuple *p = mktuple();

	p->oarg = ST;
	p->varg = v;
	p->targ[0] = t;
	return(p);
}

tuple *
stxop(var *v, tuple *t1, tuple *t2)
{
	register tuple *p = mktuple();

	p->oarg = STX;
	p->varg = v;
	p->targ[0] = t1;
	p->targ[1] = t2;
	return(p);
}

tuple *
labop(label l)
{
	register tuple *p = mktuple();

	p->oarg = LAB;
	p->larg[0] = l;
	return(p);
}

tuple *
selop(tuple *t, label l1, label l2)
{
	register tuple *p = mktuple();

	p->oarg = SEL;
	p->targ[0] = t;
	p->larg[0] = l1;
	p->larg[1] = l2;
	return(p);
}

tuple *
killop(var *v)
{
	register tuple *p = mktuple();

	p->oarg = KILL;
	p->varg = v;
	return(p);
}

void
codegen(void)
{
	/* print listing of generated code */
	register tuple *p = code.next;

	while (p != &code) {
		printf("%d\t", ttoi(p));

		switch (p->oarg) {
		case ADD:
			cgbin("add", p);
			break;
		case SUB:
			cgbin("sub", p);
			break;
		case AND:
			cgbin("and", p);
			break;
		case OR:
			cgbin("or", p);
			break;
		case XOR:
			cgbin("xor", p);
			break;
		case GT:
			cgbin("gt", p);
			break;
		case GE:
			cgbin("ge", p);
			break;
		case EQ:
			cgbin("eq", p);
			break;
		case SSL:
			cgbin("ssl", p);
			break;
		case SSR:
			cgbin("ssr", p);
			break;
		case CONST:
			printf("const(%d)", p->carg);
			break;
		case LD:
			printf("ld(%s{%d,%d})", (p->varg)->text,
				(p->varg)->deflev, (p->varg)->defblk);
			break;
		case LDX:
			printf("ldx(%s{%d,%d}, %d)", (p->varg)->text,
				(p->varg)->deflev, (p->varg)->defblk,
				ttoi(p->targ[0]));
			break;
		case ST:
			printf("st(%s{%d,%d}, %d)", (p->varg)->text,
				(p->varg)->deflev, (p->varg)->defblk,
				ttoi(p->targ[0]));
			break;
		case STX:
			printf("stx(%s{%d,%d}, %d, %d)", (p->varg)->text,
				(p->varg)->deflev, (p->varg)->defblk,
				ttoi(p->targ[0]), ttoi(p->targ[1]));
			break;
		case LAB:
			printf("lab(%d)", p->larg[0]);
			break;
		case SEL:
			printf("sel(%d, %d, %d)", ttoi(p->targ[0]),
				p->larg[0], p->larg[1]);
			break;
		case KILL:
			printf("kill(%s{%d,%d})", (p->varg)->text,
				(p->varg)->deflev, (p->varg)->defblk);
			break;
		default:
			printf("{bad opcode %d}", p->oarg);
		}

		printf("\n");
		p = p->next;
	}
}
